def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i+n]

def byteswap(data, word_size=4):
    """ Swap the byte-ordering in a packet with N=4 bytes per word
    """
    return reduce(lambda x,y: x+''.join(reversed(y)), chunks(data, word_size), '')

from construct import Construct
class PrintContext(Construct):
    def _parse(self, *args):
        _, context = args
        import pdb; pdb.set_trace()
        print context

    def _build(self, *args):
        import pdb; pdb.set_trace()



# Packet object that manages the formats of our packets
# http://stackoverflow.com/questions/5189232/how-to-auto-register-a-class-when-its-defined

from zope.interface import implements
from interfaces import IIPBusPacket
from utilities import byteswap

from constructs.ipbus import IPBusConstruct

class IPBusPacket(object):
    implements(IIPBusPacket)

    def __init__(self, blob):
        self.request = None
        self.response = None
        self._raw = blob
        # if little-endian, we need to swap when reading and writing
        self.littleendian = bool((ord(self._raw[0])&0xf0)>>4 == 0xf)
        # do some flipping
        raw = self.raw
        if self.littleendian: raw = byteswap(raw)
        self.request = IPBusConstruct.parse(raw)
        self.response = IPBusConstruct.parse(raw)

    @property
    def raw(self):
        return self._raw

    @property
    def protocol_version(self):
        return self.request.header.protocol_version

    @property
    def reserved(self):
        return self.request.header.reserved

    @property
    def packet_id(self):
        return self.request.header.id

    @property
    def byteorder(self):
        return self.request.header.byteorder

    @property
    def packet_type(self):
        return self.request.header.type_id

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.request == other.request)

    def __ne__(self, other):
        return not self.__eq__(other)



"""
    This file implements all of the various communications one might need to do.

    Jarvis provides a callback structure that looks up (in its registry) for an appropriate communication protocol.
"""

from zope.interface import implements
from ironman.interfaces import ICommunicationSlave, ICommunicationDriver
from ironman.constructs.ipbus import IPBusWords

class Jarvis(object):
    """ This is the general communication slave.

        Jarvis is what lets us pass around communications to various routes/protocols
        while keeping the details separated from us. Here's an example of how one might use it:

        >>> from ironman.communicator import Jarvis, SimpleIO
        >>> # create a Jarvis instance to manage what we want to register
        >>> j = Jarvis()
        >>> # tell Jarvis to register this class for the given route
        >>> @j.register('fpgaOne')
        ... class FileOne(SimpleIO):
        ...     __f__ = '/path/to/fileOne'
        ...
        >>> # tell Jarvis to register this class for the given route
        >>> @j.register('fpgaTwo')
        ... class FileTwo(SimpleIO):
        ...     __f__ = '/path/to/fileTwo'
        ...
        >>> # print the available registered classes
        >>> import pprint
        >>> pprint.pprint(j.registry)
        {'fpgaOne': <class 'ironman.communicator.FileOne'>,
         'fpgaTwo': <class 'ironman.communicator.FileTwo'>}

        Jarvis does the wrapping for :func:`Jarvis.register` so that a class defined at run-time is automatically inserted.

    """

    implements(ICommunicationSlave)

    def __init__(self):
        self.registry = {}

    def set_hardware_manager(self, hwmanager):
        self.hwmanager = hwmanager

    def parse_address(self, address):
        return self.hwmanager.get_route(address)

    def __call__(self, packet):
        """
            Handle CONTROL packets
        """
        if packet.request.header.type_id == 'CONTROL':
            for transaction, response in zip(packet.request.data, packet.response.data):
                response.data = transaction_response = self.__transaction__(transaction)
        return packet

    def __transaction__(self, transaction):
        protocol = self.registry.get(self.parse_address(transaction.address), None)
        if protocol is None:
            raise KeyError(transaction.address)
        protocol = protocol()
        if transaction.type_id == 'READ':
            return IPBusWords.parse(protocol.read(transaction.address, transaction.words)).data
        elif transaction.type_id == 'WRITE':
            protocol.write(transaction.address, transaction.data)
            return

    def register(self, route):
        if route is None:
            raise ValueError('Must specify a route')
        if route in self.registry:
            raise KeyError("{0:s} is an existing route to {1:s}".format(route, self.registry[route].__name__))
        def register_wrapper(cls):
            """ Duck-typing checks
            """
            getattr(cls, 'read')
            getattr(cls, 'write')
            self.registry[route] = cls
        return register_wrapper

    def unregister(self, route):
        del self.registry[route]

class SimpleIO(object):
    implements(ICommunicationDriver)
    __f__ = None

    def read(self, offset, size):
        with open(self.__f__, 'rb') as f:
            f.seek(offset)
            return f.read(4*size)

    def write(self, offset, data):
        with open(self.__f__, 'r+b') as f:
            f.seek(offset)
            return f.write(data)

class ComplexIO(object):
    implements(ICommunicationDriver)
    __f__ = {}

    def read(self, offset, size):
        with open(self.__f__.get(offset), 'rb') as f:
            return f.read(4*size)

    def write(self, offset, data):
        with open(self.__f__.get(offset), 'r+b') as f:
            return f.write(data)



from zope.interface import implements
from interfaces import IHistory

from constructs.ipbus import IPBusConstruct
import collections

class History(dict):
    implements(IHistory)

    def __init__(self, maxlen=100):
        self.maxlen = maxlen
        self.packets = collections.deque([None]*self.maxlen, maxlen=self.maxlen)

    def record(self, packet):
        # make sure packet doesn't exist first
        if packet in self.packets:
            raise ValueError("This packet has already been recorded in the history.")
        # Attempt to delete from history only if we are removing from queue
        if self.packets[0] is not None:
            del self[self.packets[0].request.header.id]
        # Add new packet to history
        self.packets.append(packet)
        self[packet.request.header.id] = (IPBusConstruct.build(packet.request), IPBusConstruct.build(packet.response))



from construct import Array, BitField, BitStruct, Embed, Enum, GreedyRange, Struct, Switch, UBInt32, SBInt32, OneOf, Pass
from ..globals import IPBUS_VERSION

from ironman.utilities import PrintContext

PacketHeaderStruct = BitStruct("header",
                        OneOf(BitField("protocol_version", 4), [IPBUS_VERSION]),
                        OneOf(BitField("reserved", 4), [0x0]),
                        BitField("id", 16),
                        OneOf(BitField("byteorder", 4), [0xf]),
                        Enum(BitField("type_id", 4),
                            CONTROL = 0x0,
                            STATUS = 0x1,
                            RESEND = 0x2
                        )
)
"""
Struct detailing the Packet Header logic
"""

ControlHeaderStruct = BitStruct("transaction",
                        OneOf(BitField("protocol_version", 4), [IPBUS_VERSION]),
                        BitField("id", 12),
                        BitField("words", 8),
                        Enum(BitField("type_id", 4),
                            READ = 0x0,
                            NOINCREAD = 0x2,
                            WRITE = 0x1,
                            NOINCWRITE = 0x3,
                            RMWBITS = 0x4,
                            RMWSUM = 0x5,
                            RCONFIG = 0x6,
                            WCONFIG = 0x7
                        ),
                        Enum(BitField("info_code", 4),
                            SUCCESS = 0x0,
                            BADHEADER = 0x1,
                            RBUSERROR = 0x4,
                            WBUSERROR = 0x5,
                            RBUSTIMEOUT = 0x6,
                            WBUSTIMEOUT = 0x7,
                            REQUEST = 0xf
                        )
)
"""
Struct detailing the Control Header logic
"""

ReadStruct = Struct("read",
                Switch("data", lambda ctx: ctx.info_code,
                    {
                        "REQUEST": Pass,
                        "SUCCESS": Array(lambda ctx: ctx.words, UBInt32("data"))
                    },
                    default = Pass
                )
)
"""
Struct detailing the Read Transaction logic
"""

WriteStruct = Struct("write",
                Switch("data", lambda ctx: ctx.info_code,
                    {
                        "REQUEST": Array(lambda ctx: ctx.words, UBInt32("data")),
                        "SUCCESS": Pass
                    },
                    default = Pass
                )
)
"""
Struct detailing the Write Transaction logic
"""

RMWBitsStruct = Struct("rmwbits",
                Switch("data", lambda ctx: ctx.info_code,
                    {
                        "REQUEST": Embed(Struct("data", UBInt32("and"), UBInt32("or"))),
                        "SUCCESS": UBInt32("data")
                    },
                    default = Pass
                )
)
"""
Struct detailing the RMWBits Transaction logic

Should compute via :math:`X \Leftarrow (X\wedge A)\\vee (B\wedge(!A))`
"""

RMWSumStruct = Struct("rmwsum",
                Switch("data", lambda ctx: ctx.info_code,
                    {
                        "REQUEST": SBInt32("addend"),  # note: signed 32-bit for subtraction!
                        "SUCCESS": UBInt32("data")
                    },
                    default = Pass
                )
)
"""
Struct detailing the RMWSum Transaction logic

Should compute via :math:`X \Leftarrow X+A`
"""

ControlStruct = Struct("ControlTransaction",
                    Embed(ControlHeaderStruct),
                    UBInt32("address"),
                    Embed(Switch("data", lambda ctx: ctx.type_id,
                        {
                            "READ": ReadStruct,
                            "NOINCREAD": ReadStruct,
                            "WRITE": WriteStruct,
                            "NOINCWRITE": WriteStruct,
                            "RMWBITS": RMWBitsStruct,
                            "RMWSUM": RMWSumStruct,
                            "RCONFIG": ReadStruct,
                            "WCONFIG": WriteStruct
                        }
                    ))
)
"""
Struct detailing the Control Action logic
"""

StatusRequestStruct = Struct("StatusTransaction", Array(15, OneOf(UBInt32("data"), [0])))
StatusResponseStruct = Struct("StatusTransaction", Array(15, UBInt32("data")))
"""
Struct detailing the Status Action logic
"""

ResendStruct = Struct("ResendTransaction")
"""
Struct detailing the Resend Action logic
"""

IPBusWords = Struct('IPBusWords', GreedyRange(UBInt32('data')))

IPBusConstruct = Struct("IPBusPacket",
                    PacketHeaderStruct,  # defined as 'header' in context
                    Switch("data", lambda ctx: ctx.header.type_id,
                        {
                            "CONTROL" : GreedyRange(ControlStruct),
                            "STATUS" : StatusRequestStruct,
                            "RESEND" : ResendStruct,
                        }
                    )
)
"""
Top-level IPBus Construct which is a packet parser/builder
"""




# Server Class managing requests over udp
from twisted.internet.protocol import DatagramProtocol
from twisted.internet.protocol import ServerFactory, Protocol, DatagramProtocol

class UDP(DatagramProtocol):
    def __init__(self, dgen):
        self.d = dgen

    def datagramReceived(self, datagram, address):
        """
        After receiving a datagram, generate the deferreds and add myself to it.
        """
        def write(result):
            print "Writing %r" % result
            self.transport.write(result, address)

        d = self.d()
        #d.addCallbacks(write, log.err)
        d.addCallback(write)  # errors are silently ignored!
        d.callback(datagram)

class TCP(Protocol):
    def __init__(self, dgen):
        self.d = dgen

    def dataReceived(self, data):
        """
        After receiving the data, generate the deferreds and add myself to it.
        """
        def write(result):
            print "Writing %r" % result
            self.transport.write(result)
        d = self.d()
        d.addCallback(write)  # errors are silently ignored!
        d.callback(data)

class TCPFactory(ServerFactory):
    protocol = TCP
    def __init__(self, dgen):
        self.d = dgen

    def buildProtocol(self, addr):
        return self.protocol(self.d)


def ServerFactory(proto, dgen):
    if proto == 'udp':
        return UDP(dgen)
    elif proto == 'tcp':
        return TCPFactory(dgen)
    else:
        return None
